@@ -1,16 +1,21 @@
+import contextlib
 import os
 import smtplib
 from unittest.mock import patch
 
+from django.contrib import admin
 from django.contrib.auth import get_user_model
 from django.contrib.auth.models import Permission
 from django.core import mail
 from django.core.exceptions import ValidationError
+from django.db import DEFAULT_DB_ALIAS
 from django.test import TestCase
 from django.urls import reverse
 from swapper import load_model
 
+from ..admin import OrganizationOwnerAdmin
 from ..apps import logger as apps_logger
+from ..multitenancy import MultitenantAdminMixin
 from .utils import (
     TestMultitenantAdminMixin,
     TestOrganizationMixin,
@@ -30,6 +35,15 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
     """ test admin site """
 
     app_label = 'openwisp_users'
+    is_integration_test = False
+
+    def assertNumQueries(self, num, using=DEFAULT_DB_ALIAS, func=None, *args, **kwargs):
+        # in integration tests the amount of queries
+        # is different so we skip this test for now
+        if self.is_integration_test:
+            return contextlib.suppress()
+        else:
+            return super().assertNumQueries(num, func=None, *args, **kwargs)
 
     def _get_org_edit_form_inline_params(self, user, organization):
         """
@@ -106,20 +120,15 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         params.pop('last_login')
         params = self._additional_params_pop(params)
         # inline emails
+        params.update(self.add_user_inline_params)
         params.update(
             {
                 'emailaddress_set-TOTAL_FORMS': 1,
                 'emailaddress_set-INITIAL_FORMS': 1,
-                'emailaddress_set-MIN_NUM_FORMS': 0,
-                'emailaddress_set-MAX_NUM_FORMS': 0,
                 'emailaddress_set-0-verified': True,
                 'emailaddress_set-0-primary': True,
                 'emailaddress_set-0-id': user.emailaddress_set.first().id,
                 'emailaddress_set-0-user': user.id,
-                f'{self.app_label}_organizationuser-TOTAL_FORMS': 0,
-                f'{self.app_label}_organizationuser-INITIAL_FORMS': 0,
-                f'{self.app_label}_organizationuser-MIN_NUM_FORMS': 0,
-                f'{self.app_label}_organizationuser-MAX_NUM_FORMS': 0,
             }
         )
         params.update(self._get_user_edit_form_inline_params(user, self._get_org()))
@@ -128,7 +137,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             params,
             follow=True,
         )
-        self.assertNotContains(response, 'error')
+        self.assertNotContains(response, 'Please correct the error below.')
         user = User.objects.get(username='changemailtest')
         email_set = user.emailaddress_set
         self.assertEqual(email_set.count(), 2)
@@ -149,18 +158,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             company='',
             location='',
         )
-        params.update(
-            {
-                'emailaddress_set-TOTAL_FORMS': 0,
-                'emailaddress_set-INITIAL_FORMS': 0,
-                'emailaddress_set-MIN_NUM_FORMS': 0,
-                'emailaddress_set-MAX_NUM_FORMS': 0,
-                f'{self.app_label}_organizationuser-TOTAL_FORMS': 0,
-                f'{self.app_label}_organizationuser-INITIAL_FORMS': 0,
-                f'{self.app_label}_organizationuser-MIN_NUM_FORMS': 0,
-                f'{self.app_label}_organizationuser-MAX_NUM_FORMS': 0,
-            }
-        )
+        params.update(self.add_user_inline_params)
         params.update(self._get_user_edit_form_inline_params(admin, self._get_org()))
         response = self.client.post(
             reverse(f'admin:{self.app_label}_user_change', args=[admin.pk]), params
@@ -449,20 +447,15 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         params.pop('_password')
         params.pop('last_login')
         params = self._additional_params_pop(params)
+        params.update(self.add_user_inline_params)
         params.update(
             {
                 'emailaddress_set-TOTAL_FORMS': 1,
                 'emailaddress_set-INITIAL_FORMS': 1,
-                'emailaddress_set-MIN_NUM_FORMS': 0,
-                'emailaddress_set-MAX_NUM_FORMS': 0,
                 'emailaddress_set-0-verified': True,
                 'emailaddress_set-0-primary': True,
                 'emailaddress_set-0-id': user.emailaddress_set.first().id,
                 'emailaddress_set-0-user': user.id,
-                f'{self.app_label}_organizationuser-TOTAL_FORMS': 0,
-                f'{self.app_label}_organizationuser-INITIAL_FORMS': 0,
-                f'{self.app_label}_organizationuser-MIN_NUM_FORMS': 0,
-                f'{self.app_label}_organizationuser-MAX_NUM_FORMS': 0,
             }
         )
         params.update(self._get_user_edit_form_inline_params(user, self._get_org()))
@@ -512,7 +505,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         path = reverse(f'admin:{self.app_label}_user_change', args=[user.pk])
         r = self.client.post(path, params, follow=True)
         self.assertEqual(r.status_code, 200)
-        self.assertNotContains(r, 'error')
+        self.assertNotContains(r, 'Please correct the error below.')
         user.refresh_from_db()
         self.assertEqual(user.groups.count(), 1)
         self.assertEqual(user.groups.get(name='Administrator').pk, group.pk)
@@ -576,7 +569,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         path = reverse(f'admin:{self.app_label}_user_change', args=[user2.pk])
         r = self.client.post(path, params, follow=True)
         self.assertEqual(r.status_code, 200)
-        self.assertNotContains(r, 'error')
+        self.assertNotContains(r, 'Please correct the error below.')
         user2.refresh_from_db()
         self.assertEqual(user2.username, 'newuser1')
 
@@ -631,7 +624,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
                     'is_staff': True,
                 },
                 'is_owner': True,
-            },
+            }
         }
         self._test_change(options)
 
@@ -644,7 +637,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
                     'is_staff': True,
                 },
                 'is_owner': False,
-            },
+            }
         }
         self._test_change(options)
 
@@ -756,7 +749,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             reverse(f'admin:{self.app_label}_user_add'), params, follow=True
         )
         self.assertEqual(res.status_code, 200)
-        self.assertNotContains(res, 'error')
+        self.assertNotContains(res, 'Please correct the error below.')
         user = User.objects.filter(username='testadd')
         self.assertEqual(user.count(), 1)
 
@@ -792,7 +785,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             reverse(f'admin:{self.app_label}_user_add'), params, follow=True
         )
         self.assertNotContains(res, 'errors field-organization')
-        self.assertNotContains(res, 'errors')
+        self.assertNotContains(res, 'Please correct the errors below.')
         queryset = User.objects.filter(username='testadd')
         self.assertEqual(queryset.count(), 1)
         user = queryset.first()
@@ -1296,6 +1289,78 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             self.assertEqual(r.status_code, 200)
             self.assertContains(r, '-DELETE">Delete')
 
+    def test_delete_org_user(self):
+        self.client.force_login(self._get_admin())
+        user1 = self._create_user(username='user1', email='user1@email.com')
+        org1 = self._create_org(name='org1')
+        org_user = self._create_org_user(user=user1, organization=org1, is_admin=True)
+
+        with self.subTest('test delete org user which belongs to owner'):
+            post_data = {'post': 'yes'}
+            url = reverse(
+                f'admin:{self.app_label}_organizationuser_delete', args=[org_user.pk]
+            )
+            r = self.client.post(url, post_data, follow=True)
+            qs = OrganizationUser.objects.filter(organization=org1, user=user1)
+            self.assertEqual(r.status_code, 200)
+            msg = (
+                't delete this organization user because it '
+                'belongs to an organization owner'
+            )
+            self.assertContains(r, msg)
+            self.assertEqual(qs.count(), 1)
+
+        with self.subTest('test delete org user which belongs to no owner'):
+            org2 = self._create_org(name='org2')
+            org_u = self._create_org_user(user=user1, organization=org2, is_admin=False)
+            post_data = {'post': 'yes'}
+            url = reverse(
+                f'admin:{self.app_label}_organizationuser_delete', args=[org_u.pk]
+            )
+            r = self.client.post(url, post_data, follow=True)
+            qs = OrganizationUser.objects.filter(organization=org2, user=user1)
+            self.assertEqual(r.status_code, 200)
+            self.assertContains(r, 'was deleted successfully.')
+            self.assertEqual(qs.count(), 0)
+
+        with self.subTest("Can not delete only owner's org user with action"):
+            post_data = {
+                'action': 'delete_selected_overridden',
+                '_selected_action': [org_user.pk],
+                'post': 'yes',
+            }
+            url = reverse(f'admin:{self.app_label}_organizationuser_changelist')
+            with self.assertNumQueries(8):
+                r = self.client.post(url, post_data, follow=True)
+            qs = OrganizationUser.objects.filter(user=user1, organization=org1)
+            self.assertEqual(r.status_code, 200)
+            self.assertContains(r, 't delete organization users which belong to owners')
+            self.assertEqual(qs.count(), 1)
+
+        with self.subTest('delete org users with some belonging to owners'):
+            org2 = self._create_org(name='org2')
+            org_user2 = self._create_org_user(user=user1, organization=org2)
+            post_data = {
+                'action': 'delete_selected_overridden',
+                '_selected_action': [org_user.pk, org_user2.pk],
+            }
+            url = reverse(f'admin:{self.app_label}_organizationuser_changelist')
+            r = self.client.post(url, post_data, follow=True)
+            self.assertEqual(r.status_code, 200)
+            msg = (
+                't delete 1 organization user because '
+                'it belongs to an organization owner'
+            )
+            self.assertContains(r, msg)
+            post_data.update({'post': 'yes'})
+            with self.assertNumQueries(17):
+                r = self.client.post(url, post_data, follow=True)
+            qs = OrganizationUser.objects.filter(pk__in=[org_user.pk, org_user2.pk])
+            self.assertEqual(r.status_code, 200)
+            self.assertContains(r, 'Successfully deleted 1 organization user.')
+            self.assertEqual(qs.count(), 1)
+            self.assertEqual(qs.first().organization, org1)
+
     @patch('sys.stdout', devnull)
     @patch('sys.stderr', devnull)
     def test_admin_add_user_with_invalid_email(self):
@@ -1318,6 +1383,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
 
     @classmethod
     def tearDownClass(cls):
+        super().tearDownClass()
         devnull.close()
 
 
@@ -1330,6 +1396,7 @@ class TestBasicUsersIntegration(
     """
 
     app_label = 'openwisp_users'
+    is_integration_test = True
 
     def _get_user_edit_form_inline_params(self, user, organization):
         organization_user = OrganizationUser.objects.get(
@@ -1374,10 +1441,44 @@ class TestBasicUsersIntegration(
             params,
             follow=True,
         )
-        self.assertNotContains(response, 'error')
+        self.assertNotContains(response, 'Please correct the error below.')
         user.refresh_from_db()
         self.assertEqual(user.bio, params['bio'])
 
+    def _delete_inline_org_user(self, is_admin=False):
+        admin = self._create_admin()
+        user = self._create_user()
+        org = self._create_org(name='inline-org')
+        self._create_org_user(organization=org, user=user, is_admin=is_admin)
+        self.client.force_login(admin)
+        params = user.__dict__
+        params.pop('phone_number')
+        params.pop('_password')
+        params.pop('last_login')
+        params = self._additional_params_pop(params)
+        params.update(self._get_user_edit_form_inline_params(user, org))
+        params.update({f'{self.app_label}_organizationuser-0-DELETE': 'on'})
+        path = reverse(f'admin:{self.app_label}_user_change', args=[user.pk])
+
+        r = self.client.post(path, params, follow=True)
+        qs = OrganizationUser.objects.filter(user=user)
+        self.assertEqual(r.status_code, 200)
+        if is_admin:
+            single_msg = (
+                't delete 1 organization user because it '
+                'belongs to an organization owner.'
+            )
+            self.assertContains(r, single_msg)
+            self.assertEqual(qs.count(), 1)
+        else:
+            self.assertEqual(qs.count(), 0)
+
+    def test_delete_inline_org_user(self):
+        self._delete_inline_org_user()
+
+    def test_delete_inline_owner_org_user(self):
+        self._delete_inline_org_user(is_admin=True)
+
 
 class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, TestCase):
     app_label = 'openwisp_users'
@@ -1393,7 +1494,7 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
         user23 = self._create_user(
             username='user23', email='user23j@something.com', is_superuser=True
         )
-        user3 = self._create_user(username='user3', email='user3@something.com',)
+        user3 = self._create_user(username='user3', email='user3@something.com')
         organization_user1 = self._create_org_user(
             organization=org1, user=user1, is_admin=True
         )
@@ -1492,3 +1593,17 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
             hidden=[user1.username],
             visible=[staff.username],
         )
+
+    def test_class_attr_regression(self):
+        class TestAdmin(MultitenantAdminMixin):
+            multitenant_parent = 'test'
+
+        owner_admin = OrganizationOwnerAdmin(Organization, admin.site)
+
+        test_admin = TestAdmin()
+        with self.subTest('multitenant_parent added to multitenant_shared_relations'):
+            self.assertIn('test', test_admin.multitenant_shared_relations)
+        with self.subTest('mutable data structure of sibling class unaffected'):
+            self.assertNotIn('test', owner_admin.multitenant_shared_relations)
+        with self.subTest('original attribute unaffected'):
+            self.assertIsNone(MultitenantAdminMixin.multitenant_shared_relations)
